import neurograd as ng
from neurograd import xp
from neurograd.nn.module import Module
from neurograd.functions.normalize import BatchNormalizer

@ng.fuse
def exp_mov_avg(old, new, momentum): # ndarrays
    # new_running = momentum*old + (1-momentum)*new
    return momentum * old + (1.0 - momentum) * new


class BatchNorm(Module):
    """
    BN for inputs shaped (N, C)
    """
    def __init__(self, num_features: int, batch_momentum: float = 0.9, epsilon: float = 1e-5):
        super().__init__()
        self.num_features = int(num_features)
        self.batch_momentum = float(batch_momentum)
        self.epsilon = float(epsilon)
        # Learnable scale/shift
        self.add_parameter("mean_scaler", ng.zeros((1, num_features), dtype=ng.float32, requires_grad=True)) # beta
        self.add_parameter("std_scaler",  ng.ones((1, num_features), dtype=ng.float32, requires_grad=True))  # gamma
        # Running stats (buffers, no grad) - store variance, not std
        self.add_buffer("running_mean", ng.zeros((1, num_features), dtype=ng.float32, requires_grad=False))
        self.add_buffer("running_var",  ng.ones((1, num_features), dtype=ng.float32, requires_grad=False))
        # Pre-create the op with the right axes
        self._bn_op = BatchNormalizer(axis=(0,), eps=self.epsilon)

    def forward(self, X):
        from neurograd import Tensor
        if self.training:
            out = self._bn_op(X, self.mean_scaler, self.std_scaler)
            # Get batch variance (without epsilon) from the normalizer
            batch_var = self._bn_op.x_std**2 - self.epsilon
            # Update running stats (on .data to avoid autograd tracking)
            self.running_mean.data[:] = exp_mov_avg(self.running_mean.data, self._bn_op.x_mean, self.batch_momentum)
            self.running_var.data[:]  = exp_mov_avg(self.running_var.data, batch_var, self.batch_momentum)
        else:
            # Compute std from running_var and epsilon for use with _affine
            running_std = xp.sqrt(self.running_var.data + self.epsilon)
            out = self._bn_op._affine(X.data, self.running_mean.data, 
                                      running_std, self.mean_scaler.data, self.std_scaler.data)
            return Tensor(out)
        return out


class BatchNorm2D(Module):
    """
    BN for NCHW inputs (N, C, H, W)
    """
    def __init__(self, num_features: int, batch_momentum: float = 0.9, epsilon: float = 1e-5):
        super().__init__()
        self.num_features = int(num_features)
        self.batch_momentum = float(batch_momentum)
        self.epsilon = float(epsilon)
        shape = (1, num_features, 1, 1)
        # Learnable scale/shift
        self.add_parameter("mean_scaler", ng.zeros(shape, dtype=ng.float32, requires_grad=True))  # beta
        self.add_parameter("std_scaler",  ng.ones(shape, dtype=ng.float32, requires_grad=True))   # gamma
        # Running stats - store variance, not std
        self.add_buffer("running_mean", ng.zeros(shape, dtype=ng.float32, requires_grad=False))
        self.add_buffer("running_var",  ng.ones(shape, dtype=ng.float32, requires_grad=False))
        # Axes over which per-channel stats are computed: (N, H, W)
        self._bn_op = BatchNormalizer(axis=(0, 2, 3), eps=self.epsilon)
    
    def forward(self, X):
        from neurograd import Tensor
        if self.training:
            out = self._bn_op(X, self.mean_scaler, self.std_scaler)
            # Get batch variance (without epsilon) from the normalizer
            batch_var = self._bn_op.x_std**2 - self.epsilon
            # Update running stats (on .data to avoid autograd tracking)
            self.running_mean.data[:] = exp_mov_avg(self.running_mean.data, self._bn_op.x_mean, self.batch_momentum)
            self.running_var.data[:]  = exp_mov_avg(self.running_var.data, batch_var, self.batch_momentum)
        else:
            # Compute std from running_var and epsilon for use with _affine
            running_std = xp.sqrt(self.running_var.data + self.epsilon)
            out = self._bn_op._affine(X.data, self.running_mean.data, 
                                      running_std, self.mean_scaler.data, self.std_scaler.data)
            return Tensor(out)
        return out


"""
Graph visualization utilities for NeuroGrad computational graphs.
"""

import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.patches import FancyBboxPatch
import numpy as xp
from typing import Dict, List, Tuple, Set, Optional, Any
from collections import defaultdict, deque
import math


class GraphVisualizer:
    """
    A class for visualizing computational graphs with automatic scaling and layout.
    """
    
    def __init__(self, tensor):
        """
        Initialize the graph visualizer for a given tensor.
        
        Args:
            tensor: The output tensor of the computational graph
        """
        self.tensor = tensor
        self.tensors = set()
        self.functions = set()
        self.edges = defaultdict(list)
        self.levels = {}
        self.positions = {}
        
        # Collect and analyze the graph
        self._collect_graph()
        self._calculate_levels()
        self._calculate_layout()
    
    def _collect_graph(self):
        """Collect all nodes (tensors and functions) in the computational graph."""
        visited = set()
        
        def dfs(node):
            node_id = id(node)
            if node_id in visited:
                return
            visited.add(node_id)
            
            # Import here to avoid circular imports
            from ..tensor import Tensor
            
            if isinstance(node, Tensor):
                self.tensors.add(node)
                if node.grad_fn is not None:
                    self.functions.add(node.grad_fn)
                    self.edges[node.grad_fn].append(node)
                    dfs(node.grad_fn)
            else:  # Function node
                self.functions.add(node)
                if hasattr(node, 'parent_tensors'):
                    for parent in node.parent_tensors:
                        self.edges[parent].append(node)
                        dfs(parent)
        
        dfs(self.tensor)
    
    def _calculate_levels(self):
        """Calculate the level (depth) of each node using topological sort."""
        in_degree = defaultdict(int)
        
        # Calculate in-degrees
        for node in list(self.tensors) + list(self.functions):
            in_degree[node] = 0
        
        for node, children in self.edges.items():
            for child in children:
                in_degree[child] += 1
        
        # Topological sort with level calculation
        queue = deque()
        for node in list(self.tensors) + list(self.functions):
            if in_degree[node] == 0:
                queue.append(node)
                self.levels[node] = 0
        
        while queue:
            current = queue.popleft()
            for child in self.edges[current]:
                in_degree[child] -= 1
                self.levels[child] = max(self.levels.get(child, 0), self.levels[current] + 1)
                if in_degree[child] == 0:
                    queue.append(child)
    
    def _calculate_layout(self):
        """Calculate positions for all nodes using a hierarchical layout."""
        # Group nodes by level
        levels_dict = defaultdict(list)
        for node, level in self.levels.items():
            levels_dict[level].append(node)
        
        max_level = max(levels_dict.keys()) if levels_dict else 0
        
        # Calculate positions level by level
        for level in range(max_level + 1):
            nodes_at_level = levels_dict[level]
            num_nodes = len(nodes_at_level)
            
            if num_nodes == 0:
                continue
            
            # Spread nodes horizontally across the level
            y = level * 2.5  # Vertical spacing between levels
            
            if num_nodes == 1:
                x_positions = [0]
            else:
                # Calculate optimal horizontal spacing
                base_spacing = max(3.0, 8.0 / num_nodes)  # Adaptive spacing
                total_width = (num_nodes - 1) * base_spacing
                x_positions = [i * base_spacing - total_width / 2 for i in range(num_nodes)]
            
            # Sort nodes for consistent positioning
            sorted_nodes = sorted(nodes_at_level, key=lambda x: (type(x).__name__, str(x)))
            
            for i, node in enumerate(sorted_nodes):
                self.positions[node] = (x_positions[i], y)
    
    def _calculate_figure_size(self, num_nodes: int) -> Tuple[float, float]:
        """Calculate optimal figure size based on node positions and count."""
        if not self.positions:
            return (10, 8)
        
        # Get bounds
        x_coords = [pos[0] for pos in self.positions.values()]
        y_coords = [pos[1] for pos in self.positions.values()]
        
        x_range = max(x_coords) - min(x_coords)
        y_range = max(y_coords) - min(y_coords)
        
        # Base size with padding
        base_width = max(10, x_range + 4)
        base_height = max(8, y_range + 4)
        
        # Scale based on number of nodes
        if num_nodes > 20:
            scale_factor = min(2.0, 1.0 + (num_nodes - 20) * 0.02)
            base_width *= scale_factor
            base_height *= scale_factor
        
        return (base_width, base_height)
    
    def _format_tensor_info(self, tensor) -> str:
        """Format tensor information for display."""
        info_parts = []
        
        # Add name
        if hasattr(tensor, 'name') and tensor.name:
            name = tensor.name
            if len(name) > 15:
                name = name[:12] + "..."
            info_parts.append(f"Name: {name}")
        
        # Add shape
        if hasattr(tensor, 'shape'):
            shape_str = str(tensor.shape)
            if len(shape_str) > 20:
                shape_str = shape_str[:17] + "..."
            info_parts.append(f"Shape: {shape_str}")
        
        # Add gradient info
        if hasattr(tensor, 'requires_grad'):
            info_parts.append(f"Grad: {tensor.requires_grad}")
        
        return "\n".join(info_parts)
    
    def _format_function_info(self, function) -> str:
        """Format function information for display."""
        name = getattr(function, 'name', type(function).__name__)
        if len(name) > 15:
            name = name[:12] + "..."
        return name
    
    def _draw_node(self, ax, node, position: Tuple[float, float], is_tensor: bool = True):
        """Draw a single node (tensor or function) on the axes."""
        x, y = position
        
        if is_tensor:
            # Draw tensor as rectangle
            info = self._format_tensor_info(node)
            bbox = FancyBboxPatch(
                (x - 0.8, y - 0.6), 1.6, 1.2,
                boxstyle="round,pad=0.1",
                facecolor='lightblue',
                edgecolor='darkblue',
                linewidth=2
            )
            ax.add_patch(bbox)
            
            # Add text
            ax.text(x, y, info, ha='center', va='center', fontsize=8, weight='bold')
            
        else:
            # Draw function as circle
            info = self._format_function_info(node)
            circle = patches.Circle((x, y), 0.5, facecolor='lightgreen', edgecolor='darkgreen', linewidth=2)
            ax.add_patch(circle)
            
            # Add text
            ax.text(x, y, info, ha='center', va='center', fontsize=9, weight='bold')
    
    def _draw_edge(self, ax, from_pos: Tuple[float, float], to_pos: Tuple[float, float], 
                   from_tensor: bool = True, to_tensor: bool = True):
        """Draw an edge between two nodes."""
        x1, y1 = from_pos
        x2, y2 = to_pos
        
        # Calculate connection points on node boundaries
        if from_tensor:
            # From tensor (rectangle)
            if x2 > x1:
                start_x = x1 + 0.8
            elif x2 < x1:
                start_x = x1 - 0.8
            else:
                start_x = x1
            
            if y2 > y1:
                start_y = y1 + 0.6
            elif y2 < y1:
                start_y = y1 - 0.6
            else:
                start_y = y1
        else:
            # From function (circle)
            angle = math.atan2(y2 - y1, x2 - x1)
            start_x = x1 + 0.5 * math.cos(angle)
            start_y = y1 + 0.5 * math.sin(angle)
        
        if to_tensor:
            # To tensor (rectangle)
            if x1 > x2:
                end_x = x2 + 0.8
            elif x1 < x2:
                end_x = x2 - 0.8
            else:
                end_x = x2
            
            if y1 > y2:
                end_y = y2 + 0.6
            elif y1 < y2:
                end_y = y2 - 0.6
            else:
                end_y = y2
        else:
            # To function (circle)
            angle = math.atan2(y1 - y2, x1 - x2)
            end_x = x2 + 0.5 * math.cos(angle)
            end_y = y2 + 0.5 * math.sin(angle)
        
        # Draw arrow
        ax.annotate('', xy=(end_x, end_y), xytext=(start_x, start_y),
                    arrowprops=dict(arrowstyle='->', lw=1.5, color='black'))
    
    def visualize(self, figsize: Optional[Tuple[float, float]] = None, 
                  title: Optional[str] = None) -> plt.Figure:
        """
        Create a visualization of the computational graph.
        
        Args:
            figsize: Size of the figure (width, height). If None, auto-calculated
            title: Title for the graph. If None, auto-generated
            
        Returns:
            matplotlib Figure object
        """
        if not self.tensors and not self.functions:
            # Empty graph
            fig, ax = plt.subplots(figsize=(8, 6))
            ax.text(0.5, 0.5, 'Empty computational graph', ha='center', va='center', 
                    transform=ax.transAxes, fontsize=14)
            ax.set_xlim(0, 1)
            ax.set_ylim(0, 1)
            ax.axis('off')
            return fig
        
        # Calculate figure size
        num_nodes = len(self.tensors) + len(self.functions)
        if figsize is None:
            figsize = self._calculate_figure_size(num_nodes)
        
        # Create figure
        fig, ax = plt.subplots(figsize=figsize)
        
        # Draw edges first (so they appear behind nodes)
        for from_node, to_nodes in self.edges.items():
            from_pos = self.positions[from_node]
            from_is_tensor = from_node in self.tensors
            
            for to_node in to_nodes:
                to_pos = self.positions[to_node]
                to_is_tensor = to_node in self.tensors
                self._draw_edge(ax, from_pos, to_pos, from_is_tensor, to_is_tensor)
        
        # Draw nodes
        for tensor_node in self.tensors:
            self._draw_node(ax, tensor_node, self.positions[tensor_node], is_tensor=True)
        
        for function_node in self.functions:
            self._draw_node(ax, function_node, self.positions[function_node], is_tensor=False)
        
        # Set title
        if title is None:
            title = f"Computational Graph ({len(self.tensors)} tensors, {len(self.functions)} operations)"
        ax.set_title(title, fontsize=14, weight='bold', pad=20)
        
        # Set axis properties
        ax.set_aspect('equal')
        ax.axis('off')
        
        # Add padding around the graph
        if self.positions:
            x_coords = [pos[0] for pos in self.positions.values()]
            y_coords = [pos[1] for pos in self.positions.values()]
            
            x_min, x_max = min(x_coords), max(x_coords)
            y_min, y_max = min(y_coords), max(y_coords)
            
            padding = 1.5
            ax.set_xlim(x_min - padding, x_max + padding)
            ax.set_ylim(y_min - padding, y_max + padding)
        
        # Add legend
        tensor_patch = patches.Rectangle((0, 0), 1, 1, facecolor='lightblue', edgecolor='darkblue', label='Tensor')
        function_patch = patches.Circle((0, 0), 0.5, facecolor='lightgreen', edgecolor='darkgreen', label='Function')
        ax.legend(handles=[tensor_patch, function_patch], loc='upper right', bbox_to_anchor=(1, 1))
        
        plt.tight_layout()
        return fig
    
    def save(self, filename: str, **kwargs):
        """
        Save the graph visualization to a file.
        
        Args:
            filename: Path to save the image
            **kwargs: Additional arguments passed to visualize()
        """
        fig = self.visualize(**kwargs)
        fig.savefig(filename, dpi=300, bbox_inches='tight')
        plt.close(fig)
        print(f"Graph saved to: {filename}")
    
    def print_structure(self, max_depth: int = 10):
        """
        Print a text representation of the computational graph structure.
        
        Args:
            max_depth: Maximum depth to traverse (prevents infinite loops)
        """
        visited = set()
        
        def print_node(node, depth=0, prefix=""):
            if depth > max_depth or node in visited:
                return
            visited.add(node)
            
            # Import here to avoid circular imports
            from ..tensor import Tensor
            
            indent = "  " * depth
            if isinstance(node, Tensor):
                shape_str = str(node.shape) if hasattr(node, 'shape') else "unknown"
                grad_str = f", grad={node.requires_grad}" if hasattr(node, 'requires_grad') else ""
                name_str = f", name={node.name}" if hasattr(node, 'name') else ""
                print(f"{indent}{prefix}Tensor(shape={shape_str}{grad_str}{name_str})")
                
                if hasattr(node, 'grad_fn') and node.grad_fn is not None:
                    print_node(node.grad_fn, depth + 1, "├── ")
            else:
                # Function node
                name = getattr(node, 'name', type(node).__name__)
                print(f"{indent}{prefix}Function({name})")
                
                if hasattr(node, 'parent_tensors'):
                    for i, parent in enumerate(node.parent_tensors):
                        is_last = i == len(node.parent_tensors) - 1
                        child_prefix = "└── " if is_last else "├── "
                        print_node(parent, depth + 1, child_prefix)
        
        print("Computational Graph Structure:")
        print("=" * 40)
        print_node(self.tensor)
        print("=" * 40)
    
    def get_stats(self) -> Dict[str, Any]:
        """
        Get statistics about the computational graph.
        
        Returns:
            Dictionary containing graph statistics
        """
        # Count different operation types
        function_types = defaultdict(int)
        for func in self.functions:
            name = getattr(func, 'name', type(func).__name__)
            function_types[name] += 1
        
        # Calculate graph depth
        max_depth = max(self.levels.values()) if self.levels else 0
        
        stats = {
            'num_tensors': len(self.tensors),
            'num_functions': len(self.functions),
            'num_edges': sum(len(children) for children in self.edges.values()),
            'max_depth': max_depth,
            'function_types': dict(function_types)
        }
        
        return stats


# Convenience functions that work with the class
def visualize_graph(tensor, **kwargs) -> plt.Figure:
    """
    Visualize the computational graph that led to this tensor.
    
    Args:
        tensor: The output tensor of the computational graph
        **kwargs: Additional arguments passed to GraphVisualizer.visualize()
        
    Returns:
        matplotlib Figure object
    """
    visualizer = GraphVisualizer(tensor)
    return visualizer.visualize(**kwargs)


def save_graph(tensor, filename: str, **kwargs):
    """
    Save a visualization of the computational graph to file.
    
    Args:
        tensor: The output tensor of the computational graph
        filename: Path to save the image
        **kwargs: Additional arguments passed to GraphVisualizer.visualize()
    """
    visualizer = GraphVisualizer(tensor)
    visualizer.save(filename, **kwargs)


def print_graph_structure(tensor, max_depth: int = 10):
    """
    Print a text representation of the computational graph structure.
    
    Args:
        tensor: The output tensor of the computational graph
        max_depth: Maximum depth to traverse (prevents infinite loops)
    """
    visualizer = GraphVisualizer(tensor)
    visualizer.print_structure(max_depth)


def get_graph_stats(tensor) -> Dict[str, Any]:
    """
    Get statistics about the computational graph.
    
    Args:
        tensor: The output tensor of the computational graph
        
    Returns:
        Dictionary containing graph statistics
    """
    visualizer = GraphVisualizer(tensor)
    return visualizer.get_stats()


import threading


_TLS = threading.local()
if not hasattr(_TLS, "enabled"):
    _TLS.enabled = True


class no_grad:
    def __init__(self, enabled: bool = True):
        self._set = bool(enabled)
        self._prev = None

    def __enter__(self):
        self._prev = getattr(_TLS, "enabled", True)
        if self._set:
            _TLS.enabled = False
        return self

    def __exit__(self, exc_type, exc, tb):
        _TLS.enabled = self._prev
        return False


def is_grad_enabled() -> bool:
    return bool(getattr(_TLS, "enabled", True))


def set_grad_enabled(mode: bool) -> None:
    _TLS.enabled = bool(mode)




from neurograd import xp
import builtins
from .base import Function
from neurograd.nn.module import Module


# CUPY MULTI-AXIS VAR BUG SLOWDOWN FIX
def _reduce(arr, axis, reduction_func, keepdims=False, **kwargs):
    """
    Performs a reduction iteratively over a tuple of axes for optimal speed,
    especially in CuPy. This function is the FAST PATH for decomposable
    reductions like mean, sum, max, and min.
    """
    # If axis is None, reduce over all axes by converting to a tuple of all axes.
    if axis is None:
        axis = tuple(range(arr.ndim))
    # If axis is a single integer, the library's default is already optimized.
    if isinstance(axis, int):
        return reduction_func(arr, axis=axis, keepdims=keepdims, **kwargs)
    # --- Iterative Path for Tuple of Axes (The Fast Path) ---
    ndim = arr.ndim
    axes = tuple(ax if ax >= 0 else ndim + ax for ax in axis)
    axes = tuple(sorted(axes, reverse=True))
    result = arr
    for ax in axes:
        result = reduction_func(result, axis=ax, keepdims=True, **kwargs)
    if not keepdims:
        result = xp.squeeze(result, axis=axes)
    return result

def _iterative_var(arr, axis, keepdims=False):
    """
    Calculates variance correctly and quickly using the fast `_reduce` helper.
    This version includes a clamp to prevent negative variance due to
    floating-point inaccuracies (catastrophic cancellation).
    """
    arr_mean = _reduce(arr, axis, xp.mean, keepdims=True)
    arr_sq_mean = _reduce(xp.square(arr), axis, xp.mean, keepdims=True)
    var = arr_sq_mean - xp.square(arr_mean)
    # --- THE FIX ---
    # Clamp the variance to be non-negative. This is a robust way to
    # handle floating-point errors without affecting correct positive variances.
    var = xp.maximum(var, 0)
    if not keepdims:
        if axis is None:
            axis_to_squeeze = tuple(range(arr.ndim))
        elif isinstance(axis, int):
            axis_to_squeeze = (axis,)
        else: # tuple
            axis_to_squeeze = axis
        var = xp.squeeze(var, axis=axis_to_squeeze)
    return var



class Sum(Function, Module):
    name = "Sum"
    def __init__(self, axis=None, keepdims=False):
        Function.__init__(self)
        Module.__init__(self)
        self.axis = axis
        self.keepdims = keepdims
    def forward(self, x: xp.ndarray) -> xp.ndarray:
        return xp.sum(x, axis=self.axis, keepdims=self.keepdims, dtype=xp.float32)
    def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
        x = self.parent_tensors[0]
        if not x.requires_grad:
            return None
        # For Sum, we just need to broadcast grad_output to input shape
        grad = grad_output
        if self.axis is not None and not self.keepdims:
            # Add dimensions back by expanding to match what keepdims=True would give
            for ax in sorted(self.axis if isinstance(self.axis, tuple) else (self.axis,)):
                # Normalize negative axes
                ax_norm = ax if ax >= 0 else len(x.shape) + ax
                grad = xp.expand_dims(grad, axis=ax_norm)
        # Broadcast to original shape
        grad = xp.broadcast_to(grad, x.shape)
        return grad

class Mean(Function, Module):
    name = "Mean"
    def __init__(self, axis=None, keepdims=False):
        Function.__init__(self)
        Module.__init__(self)
        self.axis = axis
        self.keepdims = keepdims
    def forward(self, x: xp.ndarray) -> xp.ndarray:
        return xp.mean(x, axis=self.axis, keepdims=self.keepdims, dtype=xp.float32)
    def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
        x = self.parent_tensors[0]
        if not x.requires_grad:
            return None
        # Calculate number of elements being averaged
        if self.axis is None:
            n = x.size
        else:
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            n = 1
            for ax in axes:
                ax_norm = ax if ax >= 0 else len(x.shape) + ax
                n *= x.shape[ax_norm]
        
        # Expand grad_output if keepdims=False
        go = grad_output
        if self.axis is not None and not self.keepdims:
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            for ax in sorted(axes):
                ax_norm = ax if ax >= 0 else len(x.shape) + ax
                go = xp.expand_dims(go, axis=ax_norm)
        
        # Broadcast and scale by 1/n - avoid creating intermediate copy
        go = xp.broadcast_to(go, x.shape)
        return go / n

class Max(Function, Module):
    name = "Max"
    def __init__(self, axis=None, keepdims=False):
        Function.__init__(self)
        Module.__init__(self)
        self.axis = axis
        self.keepdims = keepdims
    def forward(self, x: xp.ndarray) -> xp.ndarray:
        # Cache with keepdims=True to avoid recomputation in backward
        self.max_vals = xp.max(x, axis=self.axis, keepdims=True)
        if self.keepdims:
            return self.max_vals
        else:
            # Convert axis to positive indices for squeezing
            if self.axis is None:
                return self.max_vals.reshape(())
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            axes = tuple(ax if ax >= 0 else x.ndim + ax for ax in axes)
            return xp.squeeze(self.max_vals, axis=axes)
    def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
        x = self.parent_tensors[0]
        if not x.requires_grad:
            return None
        # Create gradient mask for max elements, splitting ties evenly along the axis.
        # Use self.max_vals which already has keepdims=True for broadcasting
        mask = (x.data == self.max_vals).astype(x.data.dtype)
        if self.axis is None:
            mask /= xp.sum(mask)  # In-place division
        else:
            count = xp.sum(mask, axis=self.axis, keepdims=True)
            count = xp.where(count == 0, 1, count)
            mask /= count  # In-place division
        # Expand and broadcast gradient
        grad = grad_output
        if self.axis is not None and not self.keepdims:
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            for ax in sorted(axes):
                ax_norm = ax if ax >= 0 else len(x.shape) + ax
                grad = xp.expand_dims(grad, axis=ax_norm)
        grad = xp.broadcast_to(grad, x.shape)
        grad = grad * mask  # Use explicit multiplication to avoid in-place modification
        return grad

class Min(Function, Module):
    name = "Min"
    def __init__(self, axis=None, keepdims=False):
        Function.__init__(self)
        Module.__init__(self)
        self.axis = axis
        self.keepdims = keepdims
    def forward(self, x: xp.ndarray) -> xp.ndarray:
        # Cache with keepdims=True to avoid recomputation in backward
        self.min_vals = xp.min(x, axis=self.axis, keepdims=True)
        if self.keepdims:
            return self.min_vals
        else:
            # Convert axis to positive indices for squeezing
            if self.axis is None:
                return self.min_vals.reshape(())
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            axes = tuple(ax if ax >= 0 else x.ndim + ax for ax in axes)
            return xp.squeeze(self.min_vals, axis=axes)
    def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
        x = self.parent_tensors[0]
        if not x.requires_grad:
            return None
        # Create gradient mask for min elements, splitting ties evenly along the axis.
        # Use self.min_vals which already has keepdims=True for broadcasting  
        mask = (x.data == self.min_vals).astype(x.data.dtype)
        if self.axis is None:
            mask /= xp.sum(mask)  # In-place division
        else:
            count = xp.sum(mask, axis=self.axis, keepdims=True)
            count = xp.where(count == 0, 1, count)
            mask /= count  # In-place division
        # Expand and broadcast gradient
        grad = grad_output
        if self.axis is not None and not self.keepdims:
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            for ax in sorted(axes):
                ax_norm = ax if ax >= 0 else len(x.shape) + ax
                grad = xp.expand_dims(grad, axis=ax_norm)
        grad = xp.broadcast_to(grad, x.shape)
        grad = grad * mask  # Use explicit multiplication to avoid in-place modification
        return grad

class Std(Function, Module):
    name = "Std"
    def __init__(self, axis=None, keepdims=False, eps=1e-8):
        Function.__init__(self)
        Module.__init__(self)
        self.axis = axis
        self.keepdims = keepdims
        self.eps = eps
    def forward(self, x: xp.ndarray) -> xp.ndarray:
        # Use stable two-pass algorithm for variance, then sqrt
        self.mean = xp.mean(x, axis=self.axis, keepdims=True, dtype=xp.float32)
        centered = x - self.mean
        self.var = xp.mean(centered * centered, axis=self.axis, keepdims=True, dtype=xp.float32)
        # Add eps to prevent sqrt of negative values due to numerical errors
        self.std_vals = xp.sqrt(xp.maximum(self.var, 0.0) + self.eps)
        if self.keepdims:
            return self.std_vals
        else:
            # Convert axis to positive indices for squeezing
            if self.axis is None:
                return self.std_vals.reshape(())
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            axes = tuple(ax if ax >= 0 else x.ndim + ax for ax in axes)
            return xp.squeeze(self.std_vals, axis=axes)
    def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
        x = self.parent_tensors[0]
        if not x.requires_grad:
            return None
        # Use cached values to avoid recomputation
        mean = self.mean
        std = self.std_vals
        # Count elements along the reduced axes (population)
        if self.axis is None:
            n = x.size
        else:
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            n = 1
            for ax in axes:
                ax_norm = ax if ax >= 0 else x.ndim + ax
                n *= x.shape[ax_norm]
        # Avoid divide-by-zero
        std_safe = xp.where(std == 0, self.eps, std)
        base_grad = (x.data - mean) / (n * std_safe)
        # If keepdims=False, expand grad_output back along reduced axes
        go = grad_output
        if self.axis is not None and not self.keepdims:
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            for ax in sorted(axes):
                ax_norm = ax if ax >= 0 else x.ndim + ax
                go = xp.expand_dims(go, axis=ax_norm)
        # Broadcast and apply chain rule
        go = xp.broadcast_to(go, x.shape)
        base_grad = base_grad * go  # Use explicit multiplication to avoid in-place modification
        return base_grad

class Var(Function, Module):
    name = "Var"
    def __init__(self, axis=None, keepdims=False, ddof=0, eps=1e-8):
        Function.__init__(self)
        Module.__init__(self)
        self.axis = axis
        self.keepdims = keepdims
        self.ddof = ddof
        self.eps = eps
    def forward(self, x: xp.ndarray) -> xp.ndarray:
        # Use stable two-pass algorithm: var = E[(X - mean)²] instead of E[X²] - E[X]²
        mean_vals = xp.mean(x, axis=self.axis, keepdims=True, dtype=xp.float32)
        centered = x - mean_vals
        var_vals = xp.mean(centered * centered, axis=self.axis, keepdims=True, dtype=xp.float32)
        
        # Apply Bessel's correction if needed
        if self.ddof > 0:
            if self.axis is None:
                n = x.size
            else:
                axes = (self.axis,) if isinstance(self.axis, int) else self.axis
                n = 1
                for ax in axes:
                    ax_norm = ax if ax >= 0 else x.ndim + ax
                    n *= x.shape[ax_norm]
            var_vals = var_vals * n / (n - self.ddof)
        
        # Cache for backward pass
        self.mean = mean_vals
        
        # Handle keepdims
        if not self.keepdims and self.axis is not None:
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            axes = tuple(ax if ax >= 0 else x.ndim + ax for ax in axes)
            var_vals = xp.squeeze(var_vals, axis=axes)
        elif not self.keepdims and self.axis is None:
            var_vals = var_vals.reshape(())
            
        return var_vals
        
    def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
        x = self.parent_tensors[0]
        if not x.requires_grad:
            return None
        
        # Count elements along the reduced axes
        if self.axis is None:
            n = x.size
        else:
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            n = 1
            for ax in axes:
                ax_norm = ax if ax >= 0 else x.ndim + ax
                n *= x.shape[ax_norm]
        
        # Denominator for Bessel correction
        denom = builtins.max(n - self.ddof, self.eps)
        scale_factor = 2.0 / denom
        
        # Expand grad_output if keepdims=False
        go = grad_output
        if self.axis is not None and not self.keepdims:
            axes = (self.axis,) if isinstance(self.axis, int) else self.axis
            for ax in sorted(axes):
                ax_norm = ax if ax >= 0 else x.ndim + ax
                go = xp.expand_dims(go, axis=ax_norm)
        
        # Broadcast grad_output to input shape and compute gradient efficiently
        go = xp.broadcast_to(go, x.shape)
        
        # Compute gradient: d/dx var(x) = 2(x - mean)/(n - ddof)
        # Optimize to avoid creating large temporary arrays
        centered_scaled = (x.data - self.mean) * scale_factor
        grad = go * centered_scaled
        
        return grad


def sum(x, axis=None, keepdims=False):
    return Sum(axis=axis, keepdims=keepdims)(x)
def mean(x, axis=None, keepdims=False):
    return Mean(axis=axis, keepdims=keepdims)(x)
def max(x, axis=None, keepdims=False):
    return Max(axis=axis, keepdims=keepdims)(x)
def min(x, axis=None, keepdims=False):
    return Min(axis=axis, keepdims=keepdims)(x)
def std(x, axis=None, keepdims=False):
    return Std(axis=axis, keepdims=keepdims)(x)
def var(x, axis=None, keepdims=False):
    return Var(axis=axis, keepdims=keepdims)(x)


import math
import random
import os
import numpy as np
from neurograd import xp, Tensor, float32, int64
from typing import Optional, List, Tuple, Union, Callable, Dict
import glob
from collections import deque
from concurrent.futures import ThreadPoolExecutor
import cv2
import os
import psutil
import warnings


# Try to import DALI first - this determines our capabilities
try:
    import nvidia.dali as dali
    from nvidia.dali import pipeline_def, Pipeline
    import nvidia.dali.fn as fn
    import nvidia.dali.types as types
    # FIX 1: Import LastBatchPolicy from its location in older DALI versions
    from nvidia.dali.plugin.base_iterator import LastBatchPolicy
    from nvidia.dali.backend import PreallocateDeviceMemory, PreallocatePinnedMemory
    DALI_AVAILABLE = True

    class DALIGenericIterator:
        """
        A generic DALI iterator that is not tied to any specific framework.
        It iterates over a DALI pipeline and yields batches of data.
        """
        def __init__(self,
                     pipeline: Pipeline,
                     output_map: List[str],
                     last_batch_policy: LastBatchPolicy,
                     auto_reset: bool = False,
                     reader_name: Optional[str] = None,
                     prepare_first_batch: bool = True):
            self._pipeline = pipeline
            self._output_map = output_map
            self._last_batch_policy = last_batch_policy
            self._auto_reset = auto_reset
            self._reader_name = reader_name

            # Try to find reader name if not provided
            if not self._reader_name:
                readers = [op.name for op in self._pipeline.ops if "readers" in op.spec.name]
                if len(readers) == 1:
                    self._reader_name = readers[0]
                else:
                    raise ValueError(f"Could not automatically determine the reader name. "
                                     f"Found {len(readers)} readers: {readers}. Please specify 'reader_name'.")

            self._size = self._pipeline.epoch_size(self._reader_name)
            self._batch_size = self._pipeline.max_batch_size
            
            if self._last_batch_policy == LastBatchPolicy.DROP:
                self._num_batches = self._size // self._batch_size
            else:  # PARTIAL or FILL
                self._num_batches = math.ceil(self._size / self._batch_size)
            
            self._counter = 0

        def __iter__(self):
            return self

        def __len__(self):
            return self._num_batches

        def __next__(self):
            if self._counter >= self._num_batches:
                if self._auto_reset:
                    self.reset()
                raise StopIteration

            try:
                outputs = self._pipeline.run()
                self._counter += 1
                
                # FIX 2: Correctly map the tuple of outputs to the output_map keys.
                batch_dict = {key: outputs[i] for i, key in enumerate(self._output_map)}
                return [batch_dict]

            except StopIteration:
                if self._auto_reset:
                    self.reset()
                raise

        def reset(self):
            self._pipeline.reset()
            self._counter = 0

except ImportError:
    DALI_AVAILABLE = False
    warnings.warn("INFO: NVIDIA DALI not available. Falling back to OpenCV-based implementation."
                  "For maximum performance, install with: "
                  "pip install --extra-index-url https://developer.download.nvidia.com/compute/redist nvidia-dali-cuda120")

# Image file extensions
IMG_EXTS = (
    '.png', '.jpg', '.jpeg', '.bmp', '.gif',
    '.tif', '.tiff', '.webp', '.jfif', '.avif',
    '.heif', '.heic'
)


class Dataset:
    """Base dataset class for simple tensor data"""
    def __init__(self, X, y, dtype=float32):
        assert len(X) == len(y), "Mismatched input and label lengths"
        self.X = Tensor(X, dtype=dtype)
        self.y = Tensor(y, dtype=dtype)
    
    def __len__(self):
        return len(self.X)
    
    def __getitem__(self, idx):
        return self.X[idx], self.y[idx]
    
    def __iter__(self):
        for idx in range(len(self)):
            yield self[idx]
    
    def shuffle(self, seed: Optional[int] = None):
        indices = list(range(len(self)))
        rng = random.Random(seed) if seed is not None else random.Random()
        rng.shuffle(indices)
        self.X = self.X[indices]
        self.y = self.y[indices]
    
    def __repr__(self):
        return f"<Dataset: {len(self)} samples, dtype={self.X.data.dtype}>"
    
    def __str__(self):
        preview_x = self.X[:1]
        preview_y = self.y[:1]
        return (f"Dataset:\n"
                f"  Total samples: {len(self)}\n"
                f"  Input preview: {preview_x}\n"
                f"  Target preview: {preview_y}")


class ImageFolder(Dataset):
    """
    ImageFolder dataset class that handles image loading and preprocessing.
    All data loading parameters are now handled by the DataLoader.
    """
    def __init__(
        self,
        root: str,
        img_shape: tuple = None,          # (H, W)
        img_mode: str = "RGB",            # "RGB", "L", etc.
        img_normalize: bool = True,       # /255 -> float
        img_transform: callable = None,   # DALI pipeline or callable
        one_hot_targets: bool = True,     # Convert targets to one-hot encoding
        img_dtype=xp.float32,
        target_dtype=xp.int64,
        chw: bool = True,                 # return CxHxW if True, else HxWxC
    ):
        self.root = root
        self.img_shape = img_shape
        self.img_mode = img_mode
        self.img_normalize = img_normalize
        self.img_transform = img_transform
        self.one_hot_targets = one_hot_targets
        self.img_dtype = img_dtype
        self.target_dtype = target_dtype
        self.chw = chw

        self.images: List[str] = []
        self.targets: List[str] = []
        self.dont_use_mmap = is_on_network_drive(self.root)
        if self.dont_use_mmap:
            warnings.warn("ImageFolder dataset directory appears to be on a network drive / non-local storage. "
                          "Performance will be severely impacted. Consider copying data locally.", UserWarning)
        self._collect_paths()

        # Check if we have any images
        if len(self.images) == 0:
            raise ValueError(f"No images found in {root} with supported extensions: {IMG_EXTS}")

        # Stable class mapping
        self.target_names = sorted(set(self.targets))
        self.target_mapping = {name: i for i, name in enumerate(self.target_names)}
        self.num_classes = len(self.target_names)
        
        # Create one-hot mapping
        self.one_hot_mapping: Dict[int, np.ndarray] = {}
        for class_idx in range(self.num_classes):
            one_hot = np.zeros(self.num_classes, dtype=np.float32)
            one_hot[class_idx] = 1.0
            self.one_hot_mapping[class_idx] = one_hot
        
        # Convert targets to numeric labels
        self.numeric_targets = [self.target_mapping[t] for t in self.targets]

        # print(f"ImageFolder initialized: {len(self)} samples, {self.num_classes} classes")

    def get_class_name(self, class_idx: int) -> str:
        """Get class name from class index"""
        if class_idx < 0 or class_idx >= len(self.target_names):
            raise ValueError(f"Class index {class_idx} out of range [0, {len(self.target_names)-1}]")
        return self.target_names[class_idx]
    
    def get_class_index(self, class_name: str) -> int:
        """Get class index from class name"""
        if class_name not in self.target_mapping:
            raise ValueError(f"Class name '{class_name}' not found in dataset")
        return self.target_mapping[class_name]
    
    def get_one_hot(self, class_idx: int) -> np.ndarray:
        """Get one-hot encoding from class index"""
        if class_idx not in self.one_hot_mapping:
            raise ValueError(f"Class index {class_idx} not found in one-hot mapping")
        return self.one_hot_mapping[class_idx]
    
    def get_class_from_one_hot(self, one_hot: np.ndarray) -> int:
        """Get class index from one-hot encoding"""
        if one_hot.shape != (self.num_classes,):
            raise ValueError(f"One-hot encoding must have shape ({self.num_classes},), got {one_hot.shape}")
        
        class_idx = np.argmax(one_hot)
        if one_hot[class_idx] != 1.0:
            raise ValueError("One-hot encoding must have exactly one element set to 1.0")
            
        return class_idx

    def _collect_paths(self):
        """Collect image paths and their class labels"""
        if not os.path.exists(self.root) or not os.path.isdir(self.root):
            raise ValueError(f"Root directory {self.root} does not exist or is not a directory")
        
        # Method 1: Class folders (ImageNet style)
        for class_name in os.listdir(self.root):
            class_path = os.path.join(self.root, class_name)
            if not os.path.isdir(class_path):
                continue
            
            for ext in IMG_EXTS:
                pattern = os.path.join(class_path, f"*{ext}")
                for img_path in glob.glob(pattern):
                    self.images.append(img_path)
                    self.targets.append(class_name)
        
        # Method 2: If no class folders found, walk directory tree
        if not self.images:
            for r, _, files in os.walk(self.root):
                for f in files:
                    if any(f.lower().endswith(ext) for ext in IMG_EXTS):
                        p = os.path.join(r, f)
                        cls = os.path.basename(os.path.dirname(p))
                        self.images.append(p)
                        self.targets.append(cls)


    def get_dali_pipeline(self, batch_size: int, shuffle: bool = True, 
                          device: str = "cpu", num_threads: int = 4, 
                          prefetch: int = 2, seed: int = 42):
        """Create optimized DALI pipeline for this dataset (only when DALI is available)"""
        if not DALI_AVAILABLE:
            return None
            
        if isinstance(self.img_transform, Pipeline):
            return self.img_transform

        is_gpu = device == "gpu"
        h, w = self.img_shape or (224, 224)
        @pipeline_def(batch_size=batch_size, num_threads=num_threads, device_id=0 if is_gpu else None, seed=seed,
                      prefetch_queue_depth=prefetch)
        def image_pipeline():
            images, labels = fn.readers.file(files=self.images, labels=self.numeric_targets, 
                                             random_shuffle=shuffle, name="Reader",
                                             initial_fill=4096, read_ahead=True,
                                             dont_use_mmap=self.dont_use_mmap, # Crucial for network storage
                                             )
            images = fn.decoders.image(images, device="mixed" if is_gpu else "cpu", output_type=types.RGB)
            images = fn.resize(images, resize_x=w, resize_y=h, interp_type=types.INTERP_LINEAR)
            
            if self.img_transform and callable(self.img_transform):
                images = self.img_transform(images)
            
            if self.img_normalize:
                images = fn.crop_mirror_normalize(
                    images, dtype=types.FLOAT, output_layout="CHW" if self.chw else "HWC",
                    mean=[0.485 * 255, 0.456 * 255, 0.406 * 255],
                    std=[0.229 * 255, 0.224 * 255, 0.225 * 255]
                )
            else:
                if self.chw: images = fn.transpose(images, perm=[2, 0, 1])
            
            if self.one_hot_targets:
                labels = fn.one_hot(labels, num_classes=self.num_classes)

            # A DALI pipeline function should return the DataNode objects directly.
            return images, labels

        pipeline = image_pipeline()
        return pipeline

    def _apply_img_transform(self, arr: np.ndarray) -> np.ndarray:
        """Apply image transforms (OpenCV fallback path)"""
        if self.img_transform is None:
            return arr
            
        if isinstance(self.img_transform, Pipeline):
            warnings.warn("WARNING: DALI Pipeline transform provided but running in fallback numpy/OpenCV mode. "
                          "Cannot apply Pipeline to ndarray; skipping transform.", UserWarning)
            return arr
            
        try:
            out = self.img_transform(image=arr)
            if isinstance(out, dict) and "image" in out:
                return out["image"]
            return out
        except TypeError:
            pass

        try:
            return self.img_transform(arr)
        except Exception as e:
            warnings.warn(f"WARNING: img_transform callable raised an exception: {e}. Returning original image.", UserWarning)
            return arr

    def _load_image_opencv(self, path: str) -> np.ndarray:
        """Load image using OpenCV (fallback implementation)"""
        mode = (self.img_mode or "RGB").upper()
        if mode in ("L", "GRAY", "GREY", "GRAYSCALE"):
            flag = cv2.IMREAD_GRAYSCALE
        elif mode == "RGBA":
            flag = cv2.IMREAD_UNCHANGED
        else:
            flag = cv2.IMREAD_COLOR
        
        try:
            flag |= cv2.IMREAD_IGNORE_ORIENTATION
        except Exception:
            pass

        arr = cv2.imread(path, flag)
        if arr is None:
            raise ValueError(f"Failed to read image: {path}")

        if mode == "RGB" and arr.ndim == 3 and arr.shape[2] == 3:
            arr = cv2.cvtColor(arr, cv2.COLOR_BGR2RGB)
        elif mode == "RGBA" and arr.ndim == 3 and arr.shape[2] == 4:
            arr = cv2.cvtColor(arr, cv2.COLOR_BGRA2RGBA)

        if self.img_shape is not None:
            h, w = self.img_shape
            arr = cv2.resize(arr, (int(w), int(h)), interpolation=cv2.INTER_LINEAR)
        
        if arr.ndim == 2:
            arr = arr[:, :, None]
        
        if self.img_transform:
            arr = self._apply_img_transform(arr)
        
        if self.chw and arr.ndim == 3:
            arr = np.transpose(arr, (2, 0, 1))
        
        if self.img_normalize:
            arr = arr.astype(np.float32) / 255.0
        else:
            arr = arr.astype(np.float32)
        
        return arr

    def __getitem__(self, idx: int):
        """Get single item - uses OpenCV fallback"""
        img_path = self.images[idx]
        target = self.numeric_targets[idx]
        image = self._load_image_opencv(img_path)
        
        if self.one_hot_targets:
            target = self.one_hot_mapping[target]
            target_dtype = float32
        else:
            target_dtype = self.target_dtype
        
        return Tensor(image, dtype=self.img_dtype), Tensor(target, dtype=target_dtype)

    def shuffle(self, seed: Optional[int] = None):
        """Shuffle the dataset"""
        rng = random.Random(seed) if seed is not None else random.Random()
        combined = list(zip(self.images, self.targets, self.numeric_targets))
        rng.shuffle(combined)
        self.images, self.targets, self.numeric_targets = [list(t) for t in zip(*combined)]

    def __len__(self):
        return len(self.images)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __str__(self):
        return (f"ImageFolder(root='{self.root}', samples={len(self)}, "
                f"classes={self.num_classes})")

    def __repr__(self):
        shape = tuple(self[0][0].shape) if len(self) > 0 else None
        return (f"ImageFolder(root='{self.root}', samples={len(self)}, "
                f"classes={self.num_classes}, shape={shape})")


class DataLoader:
    """
    DataLoader that handles all data loading parameters, with an efficient DALI backend
    and a multithreaded OpenCV fallback.
    """
    def __init__(
        self,
        dataset: Union[ImageFolder, Dataset],
        batch_size: int = 32,
        shuffle: bool = True,
        device: str = "gpu",
        num_workers: int = None,
        prefetch_batches: int = 2,
        drop_last: bool = False,
        seed: Optional[int] = None,
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.device = "gpu" if device in ["gpu", "cuda"] else "cpu"
        self.num_workers = os.cpu_count() if num_workers is None else max(0, int(num_workers))
        self.prefetch_batches = max(0, int(prefetch_batches))
        self.drop_last = drop_last
        self.seed = seed
        
        self._pipeline = None
        self._dali_iter = None
        self._executor: Optional[ThreadPoolExecutor] = None
        
        if DALI_AVAILABLE:
            try:
                if self.device == "gpu": PreallocateDeviceMemory(int(0.5 * 1024**3), 0)
                PreallocatePinnedMemory(int(0.25 * 1024**3))
            except Exception as e: warnings.warn(f"WARNING: DALI memory preallocation failed: {e}")
        
        self.use_dali = DALI_AVAILABLE and isinstance(dataset, ImageFolder)
        if self.use_dali:
            self._init_dali_pipeline()
        
        if self.device == "gpu" and not self.use_dali:
            warnings.warn("WARNING: GPU device requested but DALI not available. Using CPU fallback.")
            self.device = "cpu"

    def _init_dali_pipeline(self):
        if not self.use_dali: return
        self._pipeline = self.dataset.get_dali_pipeline(
            batch_size=self.batch_size, shuffle=self.shuffle, device=self.device,
            num_threads=self.num_workers, prefetch=self.prefetch_batches, seed=self.seed)
        if self._pipeline:
            policy = LastBatchPolicy.DROP if self.drop_last else LastBatchPolicy.PARTIAL
            self._dali_iter = DALIGenericIterator(self._pipeline, ["images", "labels"], policy, reader_name="Reader")
            # print(f"DALI pipeline initialized: batch_size={self.batch_size}, device={self.device}, num_threads={self.num_workers}")

    def __len__(self):
        n = len(self.dataset)
        return n // self.batch_size if self.drop_last else math.ceil(n / self.batch_size)

    def __iter__(self):
        if self.use_dali and self._dali_iter:
            self._dali_iter.reset()
            return self._dali_iterator()
        return self._regular_iterator()

    def reset(self):
        """
        Resets the internal iterator, particularly for the DALI pipeline.
        This allows re-iteration over the dataset from the beginning without
        recreating the DataLoader.
        """
        if self.use_dali and self._dali_iter:
            self._dali_iter.reset()
        # For the regular iterator, a new one is created on each __iter__ call,
        # so an explicit reset is not strictly necessary in the same way.

    
    def _dali_iterator(self):
        for data in self._dali_iter:
            # The pipeline outputs DALI Tensors which are extracted from the output dictionary.
            images_tensor = data[0]["images"]  # This is a TensorListGPU containing the batch
            labels_tensor = data[0]["labels"]  # This is a TensorListGPU containing the batch
            
            # Convert DALI TensorList to proper arrays
            if self.device == 'gpu':
                import cupy as cp
                # For GPU TensorList, convert to cupy arrays
                images_array = cp.asarray(images_tensor.as_tensor())
                labels_array = cp.asarray(labels_tensor.as_tensor())
                
                X = Tensor(images_array, dtype=self.dataset.img_dtype)
                y = Tensor(labels_array, dtype=self.dataset.target_dtype if not self.dataset.one_hot_targets else float32)
            else: # CPU device
                # For CPU TensorList, convert to numpy array
                X = Tensor(images_tensor.as_array(), dtype=self.dataset.img_dtype)
                y = Tensor(labels_tensor.as_array(), dtype=self.dataset.target_dtype if not self.dataset.one_hot_targets else float32)
    
            yield X, y
            
    def _regular_iterator(self):
        """Fallback iterator with threading and prefetching."""
        batches = list(self._batch_indices())
        window = deque()
        next_to_submit = 0
        total = len(batches)

        for _ in range(min(self.prefetch_batches, total)):
            futs = self._schedule_batch(batches[next_to_submit])
            window.append(futs)
            next_to_submit += 1

        for _ in range(total):
            if not window:
                futs = self._schedule_batch(batches[next_to_submit])
                window.append(futs)
                next_to_submit += 1

            futs = window.popleft()

            if next_to_submit < total:
                next_futs = self._schedule_batch(batches[next_to_submit])
                window.append(next_futs)
                next_to_submit += 1

            yield self._gather_batch(futs)

    def _batch_indices(self):
        n = len(self.dataset)
        order = list(range(n))
        if self.shuffle:
            random.Random(self.seed).shuffle(order)
        
        limit = (n // self.batch_size) * self.batch_size if self.drop_last else n
        for start in range(0, limit, self.batch_size):
            end = min(start + self.batch_size, n)
            yield order[start:end]

    def _ensure_executor(self):
        if self.num_workers > 0 and self._executor is None:
            self._executor = ThreadPoolExecutor(max_workers=self.num_workers)

    def _schedule_batch(self, idxs):
        if self.num_workers > 0:
            self._ensure_executor()
            return [self._executor.submit(self.dataset.__getitem__, i) for i in idxs]
        return [self.dataset[i] for i in idxs]

    def _gather_batch(self, futures_or_results):
        if self.num_workers > 0:
            batch = [f.result() for f in futures_or_results]
        else:
            batch = futures_or_results
        
        Xs, ys = zip(*batch)
        return Tensor(xp.stack([x.data for x in Xs])), Tensor(xp.stack([y.data for y in ys]))

    def close(self):
        if self._executor is not None:
            self._executor.shutdown(wait=True)
            self._executor = None


def create_dali_transforms(
    device: str = "gpu",
    brightness: float = 0.0,
    contrast: float = 0.0,
    saturation: float = 0.0,
    hue: float = 0.0,
    horizontal_flip_prob: float = 0.0,
    vertical_flip_prob: float = 0.0,
    rotation_angle: float = 0.0
):
    """
    Create common DALI augmentation transforms.
    Returns None if DALI is not available.
    """
    if not DALI_AVAILABLE:
        warnings.warn("WARNING: DALI not available. Transform creation skipped.")
        return None
        
    def apply_transforms(images):
        if brightness or contrast or saturation or hue:
            images = fn.color_twist(
                images, device=device,
                brightness=fn.random.uniform(range=[-brightness, brightness]),
                contrast=fn.random.uniform(range=[1-contrast, 1+contrast]),
                saturation=fn.random.uniform(range=[1-saturation, 1+saturation]),
                hue=fn.random.uniform(range=[-hue, hue]),
            )
        if horizontal_flip_prob > 0.0:
            images = fn.flip(images, device=device, horizontal=fn.random.coin_flip(probability=horizontal_flip_prob))
        if vertical_flip_prob > 0.0:
            images = fn.flip(images, device=device, vertical=fn.random.coin_flip(probability=vertical_flip_prob))
        if rotation_angle != 0.0:
            images = fn.rotate(
                images, device=device, angle=fn.random.uniform(range=[-rotation_angle, rotation_angle]),
                keep_size=True, fill_value=0
            )
        return images
    
    return apply_transforms




def is_on_network_drive(path_to_check: str) -> bool:
    """
    Detects if the given path is on a network-mounted filesystem.
    """
    if not os.path.exists(path_to_check):
        raise FileNotFoundError(f"Path does not exist: {path_to_check}")
    NETWORK_FS_TYPES = [
        "nfs", "nfs4", "nfsd", "cifs", "smbfs", "smb", "smb2", "smb3",
        "fuse.sshfs", "fuse.gcsfuse", "fuse.s3fs"
    ]
    target_path = os.path.abspath(path_to_check)
    partitions = psutil.disk_partitions(all=True)
    # Find the most specific mount point for the path
    longest_match = None
    for p in partitions:
        if target_path.startswith(p.mountpoint):
            if longest_match is None or len(p.mountpoint) > len(longest_match.mountpoint):
                longest_match = p
    if longest_match is None:
        # Could not find any mount, assume local
        return False
    fs_type = longest_match.fstype.lower()
    if fs_type in NETWORK_FS_TYPES:
        return True
    # Overlay is usually local (Docker /tmp), so treat it as local
    return False